@@ -16,7 +16,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
- * $Id: bacon-video-widget-xine.c,v 1.214 2005/11/24 21:37:47 hadess Exp $
+ * $Id: bacon-video-widget-xine.c,v 1.225 2006/02/03 14:27:56 hadess Exp $
  *
  * the xine engine in a widget - implementation
  */
@@ -34,6 +34,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <sched.h>
 /* X11 */
 #include <X11/X.h>
 #include <X11/Xlib.h>
@@ -47,6 +48,7 @@
 
 #include "debug.h"
 #include "bacon-video-widget.h"
+#include "bacon-video-widget-private.h"
 #include "baconvideowidget-marshal.h"
 #include "video-utils.h"
 #include "bacon-resize.h"
@@ -148,6 +150,7 @@ struct BaconVideoWidgetPrivate {
 	xine_post_t *vis;
 	GList *visuals;
 	char *queued_vis;
+	VisualsQuality quality;
 
 	/* Seeking stuff */
 	int seeking;
@@ -167,6 +170,7 @@ struct BaconVideoWidgetPrivate {
 	gboolean is_live;
 	char *codecs_path;
 	gboolean got_redirect;
+	gboolean has_subtitle;
 
 	GAsyncQueue *queue;
 	int video_width, video_height;
@@ -251,6 +255,8 @@ static xine_video_port_t * load_video_out_driver (BaconVideoWidget *bvw,
 static xine_audio_port_t * load_audio_out_driver (BaconVideoWidget *bvw,
 		gboolean null_out, GError **error);
 static gboolean bacon_video_widget_tick_send (BaconVideoWidget *bvw);
+static void bacon_video_widget_set_visuals_quality_size (BaconVideoWidget *bvw,
+		int h, int w, int fps);
 
 static GtkWidgetClass *parent_class = NULL;
 
@@ -451,12 +457,17 @@ bacon_video_widget_finalize (GObject *object)
 {
 	BaconVideoWidget *bvw = (BaconVideoWidget *) object;
 
+	if (bvw->priv->gc)
+		g_object_unref (G_OBJECT (bvw->priv->gc));
+
 	if (bvw->priv->xine != NULL) {
+		xine_plugins_garbage_collector (bvw->priv->xine);
 		xine_exit (bvw->priv->xine);
 		bvw->priv->xine = NULL;
 	}
 	if (bvw->priv->cursor != NULL) {
 		gdk_cursor_unref (bvw->priv->cursor);
+		bvw->priv->cursor = NULL;
 	}
 	g_free (bvw->priv->vis_name);
 	g_free (bvw->priv->mediadev);
@@ -1043,22 +1054,27 @@ configure_cb (GtkWidget *widget, GdkEventConfigure *event,
 static void
 size_changed_cb (GdkScreen *screen, BaconVideoWidget *bvw)
 {
-	double res_h, res_v;
+	double res_h, res_v, vis_width;
 
-	XLockDisplay (bvw->priv->display);
-	res_h = (DisplayWidth (bvw->priv->display, bvw->priv->screen) * 1000 /
-			DisplayWidthMM (bvw->priv->display,
-				bvw->priv->screen));
-	res_v = (DisplayHeight (bvw->priv->display, bvw->priv->screen) * 1000 /
-			DisplayHeightMM (bvw->priv->display,
-				bvw->priv->screen));
-	XUnlockDisplay (bvw->priv->display);
+	res_h = gdk_screen_get_width (screen) * 1000 /
+		gdk_screen_get_width_mm (screen);
+	res_v = gdk_screen_get_height (screen) * 1000 /
+		gdk_screen_get_height_mm (screen);
+
+	vis_width = vis_qualities[bvw->priv->quality].height *
+		gdk_screen_get_width (screen) /
+		gdk_screen_get_height (screen);
 
 	bvw->priv->display_ratio = res_v / res_h;
 
 	if (fabs (bvw->priv->display_ratio - 1.0) < 0.01) {
 		bvw->priv->display_ratio = 1.0;
 	}
+
+	bacon_video_widget_set_visuals_quality_size (bvw,
+			vis_width,
+			vis_qualities[bvw->priv->quality].height,
+			vis_qualities[bvw->priv->quality].fps);
 }
 
 static void
@@ -1421,6 +1437,43 @@ xine_event (void *user_data, const xine_event_t *event)
 	}
 }
 
+static int
+bacon_video_widget_sort_queue (gconstpointer a, gconstpointer b, gpointer data)
+{
+	signal_data *one, *two;
+
+	one = (signal_data *) a;
+	two = (signal_data *) b;
+
+	if (one->signal == two->signal)
+		return 0;
+	if (one->signal == ERROR_ASYNC || one->signal == MESSAGE_ASYNC)
+		return -1;
+	return 1;
+}
+
+#if (!(GLIB_CHECK_VERSION(2,8,0)))
+static void
+bacon_video_widget_queue_sort (GAsyncQueue *queue)
+{
+	GList *list = NULL, *l;
+	signal_data *data;
+
+	while ((data = g_async_queue_try_pop (queue)) != NULL) {
+		list = g_list_insert_sorted (list, data, (GCompareFunc) bacon_video_widget_sort_queue);
+	}
+
+	if (list == NULL)
+		return;
+
+	for (l = list; l != NULL; l = l->next) {
+		g_async_queue_push (queue, l->data);
+	}
+	g_list_free (list);
+}
+
+#endif /* ! GLIB_CHECK_VERSION 2.8.0 */
+
 static void
 xine_try_error (BaconVideoWidget *bvw, gboolean probe_error, GError **error)
 {
@@ -1429,12 +1482,21 @@ xine_try_error (BaconVideoWidget *bvw, gboolean probe_error, GError **error)
 
 	save_data = NULL;
 
+	sched_yield ();
+
+	/* Sort the queue with the errors first */
+#if GLIB_CHECK_VERSION(2,8,0)
+	g_async_queue_sort (bvw->priv->queue, bacon_video_widget_sort_queue, NULL);
+#else
+	bacon_video_widget_queue_sort (bvw->priv->queue);
+#endif
+
 	/* Steal messages from the async queue, if there's an error,
 	 * to use as the error message rather than the crappy errors from
 	 * xine_open() */
 	while ((data = g_async_queue_try_pop (bvw->priv->queue)) != NULL)
 	{
-		if (data->signal == ERROR_ASYNC)
+		if (data->signal == ERROR_ASYNC || data->signal == MESSAGE_ASYNC)
 		{
 			if (save_data != NULL)
 			{
@@ -1443,8 +1505,8 @@ xine_try_error (BaconVideoWidget *bvw, gboolean probe_error, GError **error)
 			}
 			save_data = data;
 		} else {
-			g_free (data->msg);
-			g_free (data);
+			g_async_queue_push (bvw->priv->queue, data);
+			break;
 		}
 	}
 
@@ -1566,6 +1628,7 @@ bacon_video_widget_unrealize (GtkWidget *widget)
 				bvw->priv->ao_driver);
 
 	/* stop event thread */
+	xine_plugins_garbage_collector (bvw->priv->xine);
 	xine_exit (bvw->priv->xine);
 	bvw->priv->xine = NULL;
 
@@ -2006,11 +2069,14 @@ bacon_video_widget_open_with_subtitle (BaconVideoWidget *bvw, const char *mrl,
 		char *subtitled;
 		subtitled = bacon_video_widget_get_subtitled (mrl, subtitle_uri);
 		err = xine_open (bvw->priv->stream, subtitled);
+		bvw->priv->has_subtitle = TRUE;
 		g_free (subtitled);
 	} else {
 		err = xine_open (bvw->priv->stream, mrl);
 	}
 
+	xine_plugins_garbage_collector (bvw->priv->xine);
+
 	if (err == 0)
 	{
 		bacon_video_widget_close (bvw);
@@ -2248,6 +2314,7 @@ bacon_video_widget_close (BaconVideoWidget *bvw)
 
 	xine_stop (bvw->priv->stream);
 	xine_close (bvw->priv->stream);
+	bvw->priv->has_subtitle = FALSE;
 	g_free (bvw->priv->mrl);
 	bvw->priv->mrl = NULL;
 
@@ -2838,44 +2905,11 @@ bacon_video_widget_set_visuals (BaconVideoWidget *bvw, const char *name)
 	return FALSE;
 }
 
-void
-bacon_video_widget_set_visuals_quality (BaconVideoWidget *bvw,
-		VisualsQuality quality)
+static void
+bacon_video_widget_set_visuals_quality_size (BaconVideoWidget *bvw,
+		int w, int h, int fps)
 {
 	xine_cfg_entry_t entry;
-	int fps, w, h;
-
-	g_return_if_fail (bvw != NULL);
-	g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
-	g_return_if_fail (bvw->priv->xine != NULL);
-
-	switch (quality)
-	{
-	case VISUAL_SMALL:
-		fps = 15;
-		w = 320;
-		h = 240;
-		break;
-	case VISUAL_NORMAL:
-		fps = 25;
-		w = 320;
-		h = 240;
-		break;
-	case VISUAL_LARGE:
-		fps = 25;
-		w = 640;
-		h = 480;
-		break;
-	case VISUAL_EXTRA_LARGE:
-		fps = 30;
-		w = 800;
-		h = 600;
-		break;
-	default:
-		/* shut up warnings */
-		fps = w = h = 0;
-		g_assert_not_reached ();
-	}
 
 	bvw_config_helper_num (bvw->priv->xine, "effects.goom.fps", fps, &entry);
 	entry.num_value = fps;
@@ -2890,6 +2924,31 @@ bacon_video_widget_set_visuals_quality (BaconVideoWidget *bvw,
 	xine_config_update_entry (bvw->priv->xine, &entry);
 }
 
+void
+bacon_video_widget_set_visuals_quality (BaconVideoWidget *bvw,
+		VisualsQuality quality)
+{
+	int fps, h, w;
+
+	g_return_if_fail (bvw != NULL);
+	g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
+	g_return_if_fail (bvw->priv->xine != NULL);
+	g_return_if_fail (quality < NUM_VISUAL_QUALITIES);
+
+	fps = vis_qualities[quality].fps;
+	h = vis_qualities[quality].height;
+
+	XLockDisplay (bvw->priv->display);
+	w = vis_qualities[quality].height *
+		DisplayWidth (bvw->priv->display, bvw->priv->screen) /
+		DisplayHeight (bvw->priv->display, bvw->priv->screen);
+	XUnlockDisplay (bvw->priv->display);
+
+	bacon_video_widget_set_visuals_quality_size (bvw, w, h, fps);
+
+	bvw->priv->quality = quality;
+}
+
 gboolean
 bacon_video_widget_get_auto_resize (BaconVideoWidget *bvw)
 {
@@ -3558,8 +3617,14 @@ GList
 	num_channels = xine_get_stream_info
 		(bvw->priv->stream, XINE_STREAM_INFO_MAX_SPU_CHANNEL);
 
-	if (num_channels < 2)
+	if (num_channels < 1) {
+		if (bvw->priv->has_subtitle != FALSE) {
+			return g_list_prepend (list,
+					(gpointer) g_strdup_printf
+					(_("Language %d"), 0));
+		}
 		return NULL;
+	}
 
 	for(i = 0; i < num_channels; i++)
 	{
@@ -3610,14 +3675,6 @@ bacon_video_widget_can_get_frames (BaconVideoWidget *bvw, GError **error)
 	g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
 	g_return_val_if_fail (bvw->priv->xine != NULL, FALSE);
 
-	if (xine_get_status (bvw->priv->stream) != XINE_STATUS_PLAY
-			&& bvw->priv->logo_mode == FALSE)
-	{
-		g_set_error (error, BVW_ERROR, BVW_ERROR_CANNOT_CAPTURE,
-				_("Movie is not playing."));
-		return FALSE;
-	}
-
 	if (xine_get_stream_info (bvw->priv->stream,
 				XINE_STREAM_INFO_HAS_VIDEO) == FALSE
 			&& bvw->priv->using_vfx == FALSE)
@@ -3635,6 +3692,14 @@ bacon_video_widget_can_get_frames (BaconVideoWidget *bvw, GError **error)
 		return FALSE;
 	}
 
+	if (xine_get_status (bvw->priv->stream) != XINE_STATUS_PLAY
+			&& bvw->priv->logo_mode == FALSE)
+	{
+		g_set_error (error, BVW_ERROR, BVW_ERROR_CANNOT_CAPTURE,
+				_("Movie is not playing."));
+		return FALSE;
+	}
+
 	return TRUE;
 }
 
